# /**
#
# \name{plotDataMatrix}
#
# \title{Visualisation of a data matrix, e.g. a matrix of distance or similarity scores}
#
# \alias{plotDataMatrix}
#
# \keyword{}
#
# \description{Visualisation of a data matrix, e.g. a matrix of distance or similarity scores}
#
#
# \arguments{
#  \item{dataMatrix}{data matrix, type matrix}
#  \item{labelMatrix}{matrix of labels of type array of dim(\code{dataMatrix})}
#  \item{title}{plot title, type character string}
#  \item{labels}{vector of names corresponding to  a quadratic \code{dataMatrix}}
#  \item{xLabels}{vector of names corresponding to the columns of \code{dataMatrix} overrides \code{labels}}
#  \item{yLabels}{vector of names corresponding to the rows of \code{dataMatrix} overrides \code{labels}}
#  \item{axesLabel}{label for x and y axes}
#  \item{xAxisLabel}{label for x axis overrides \code{axesLabel}}
#  \item{yAxisLabel}{label for y axis overrides \code{axesLabel}}
#  \item{colourRamp}{vector of colours}
#  \item{minimum}{numeric, predefined minimum value, default missing}
#  \item{maximum}{numeric, predefined maximum value, default missing}
#  \item{width}{graphics window width}
#  \item{height}{graphics window height}
#  \item{labelMatrixTextScaling}{numeric, default: 0.7}
#  \item{plotOutput}{character string specifying
#          either "screen" or "pdf"; default: "screen"}
#  \item{fileName}{character string specifying the file path and file name}
# }
#
# \details{if no labels are supplied the dimnames or alternatively a numbering is used instead}
#
# \examples{plotDataMatrix(title="Similarities",dataMatrix=matrix(c(3,4,4,3),nrow=2, ncol=2),labels=c("one","two"))
#ma <- matrix(c(0.3,0.01,0.7,0.1,0.5,0.3,1,0.5,01), nrow=3,ncol=3)
# plotDataMatrix(title="Similarities",dataMatrix=ma, labelMatrix=ma,labels=c("one","two", "three"), minimum=0, maximum=1)
# ma <- matrix(c(0.3, 0.01, 0.7, 0.1, 0.5, 0.3, 1, 0.5, 1, 0, 1, 0), nrow = 4, ncol = 3, byrow=TRUE )
#plotDataMatrix(title = "dataMatrix", dataMatrix = ma, labelMatrix = ma, xLabels = c("one", "two", "three"), minimum = 0, maximum = 1)
# }
#
# \author{Andreas Buness <a.buness@dkfz.de>}
#
# */

plotDataMatrix <- function(dataMatrix, labelMatrix, title="", labels, xLabels, yLabels, axesLabel="", xAxisLabel, yAxisLabel, colourRamp, minimum, maximum, width=8, height=7, labelMatrixTextScaling=0.7, plotOutput="screen", fileName){

  if( missing(fileName) ){
    fileName <- "plotDataMatrixOutput.pdf"
  }
  
  if( ! plotOutput %in% c("screen","pdf") ){
    warning(" unknown plotOutput in plotDataMatrix", call.=FALSE)
    return()
  }
  
  if( missing(xAxisLabel) ){
    xAxisLabel <- axesLabel
  }
  if( missing(yAxisLabel) ){
    yAxisLabel <- axesLabel
  }
  
  if( (! missing( minimum ) & missing( maximum )) |
      (! missing( maximum ) & missing( minimum ))   ){
    warning(" both minimum AND maximum Or none required in plotDataMatrix !", call.=FALSE)
    return()
  }
  if( !missing(minimum) & !missing(maximum) ){
    if( !is.numeric(minimum) | !is.numeric(maximum) ){
      warning(" minimum and maximum have to be numeric in plotDataMatrix", call.=FALSE)
      return()
    }
    if( ! (minimum < maximum) ){
      warning(" minimum is not less than maximum in plotDataMatrix", call.=FALSE)
      return()
    }
  }
  
  
  if( ! is.matrix(dataMatrix) ){
    warning(" type matrix is required in plotDataMatrix", call.=FALSE)
    return()
  }
  
  nrOfRows <- dim(dataMatrix)[1]
  nrOfColumns <- dim(dataMatrix)[2]

  if( !missing(labels) ){
    if( nrOfRows == nrOfColumns ){
      if( missing( xLabels ) ){
        xLabels <- labels
      }
      if( missing( yLabels ) ){
        yLabels <- labels
      }
    }else if( nrOfRows != nrOfColumns){
      warning(" labels only useful together with quadratic dataMatrix in plotDataMatrix ", call.=FALSE)
      return()
    }
  }

  if( missing(xLabels) || all(is.null(xLabels)) || all(is.na(xLabels)) ){
    dn <- dimnames(dataMatrix)
    if( !is.null(dn[[1]]) ){
      xLabels <- dn[[1]]
    }else{
      xLabels <- as.character(1:nrOfColumns)
    }
  }
  if( missing(yLabels) || all(is.null(yLabels)) || all(is.na(yLabels)) ){
    dn <- dimnames(dataMatrix)
    if( !is.null(dn[[1]]) ){
      yLabels <- dn[[1]]
    }else{
      yLabels <- as.character(1:nrOfRows)
    }
  }
    
    
  if( !missing(labelMatrix) ){
    if( ! is.array(labelMatrix) ){
      warning("  labelMatrix must be of type array in plotDataMatrix", call.=FALSE)
       return()
    }
    if( any(dim(labelMatrix) != dim(dataMatrix)) ){
      warning("  labelMatrix and dataMatrix must have same dimension in plotDataMatrix", call.=FALSE)
      return()
    }
  }
      
  

    
  if( missing(colourRamp) ){
    colourRamp <- rgb(seq(0,1,l=256),seq(0,1,l=256),seq(1,0,l=256))
  }
  if( !missing(colourRamp) ){
    if( any(is.na(colourRamp))   | any(is.null(colourRamp)) ){
      colourRamp <- rgb(seq(0,1,l=256),seq(0,1,l=256),seq(1,0,l=256))
    }
  }

  localMinimum <- min(dataMatrix, na.rm=TRUE)
  localMaximum <- max(dataMatrix, na.rm=TRUE)

  if( !missing(minimum) & !missing(maximum) ){

    if( ! (localMinimum >= minimum && localMaximum <= maximum ) ){
      warning("data are not in the range of (minimum,maximum) in plotDataMatrix", call.=FALSE)
      return()
    }
      
    # global scale
    collevs = seq( minimum, maximum, length=length(colourRamp))
    grade <- length(colourRamp)
    gradeUnit <- (maximum - minimum)/grade
    gradeMin <- (localMinimum - minimum) %/% gradeUnit
    gradeMax <- (localMaximum - minimum) %/% gradeUnit
    if( is.nan(gradeMin) | is.nan(gradeMax) ){
      warning(" is.nan gradeMin or gradeMax in plotDataMatrix ", call.=FALSE)
      return()
    }
    if( gradeMin < 0 | gradeMax < 0 ){
      warning(" gradeMin or gradeMax < 0 in plotDataMatrix ", call.=FALSE)
      return()
    }
    if( gradeMin > length(colourRamp) | gradeMax > length(colourRamp) ){
      warning(" gradeMin or gradeMax > length(colourRamp) in plotDataMatrix ", call.=FALSE)
      return()
    }
    localColourRamp <-  colourRamp[gradeMin:gradeMax]

  }else{

    # local scale
    collevs = seq(localMinimum, localMaximum, length=length(colourRamp))
    localColourRamp <- colourRamp

  }

  if( plotOutput == "screen" ){
    x11(width=width,height=height)
  }else if( plotOutput == "pdf" ){
    pdf(width=width,height=height,file=fileName)
  }else{
    stop(" unexpected case; unknown plotOutput in plotDataMatrix ")
  }

  lo <- layout(matrix(data = c(1,2), nrow=1, ncol=2), widths=c(4, 1), heights=c(1,1))
  image(1:length(xLabels), 1:length(yLabels), t(dataMatrix), col=localColourRamp, axes=FALSE, xlab=xAxisLabel, ylab=yAxisLabel)
  axis(1,1:length(xLabels),xLabels)
  axis(2,1:length(yLabels),yLabels)
  title(main=title)
  
  if( collevs[1] < collevs[length(collevs)] ){
    image(1, collevs, matrix(data=collevs, ncol=length(collevs), nrow=1),
          col=colourRamp, xlab="", ylab="", xaxt="n" )
  }else{
    cat(paste(" no colour bar in plotDataMatrix: the minimum is not less than the maximum ( minimum:", collevs[1], " and maximum:", collevs[length(collevs)], ")",sep=""))
  }

  if( ! missing( labelMatrix ) ){

    par(mfg=c(1,1))
    par(new=TRUE)
    plot(1,1, xlim=c(1-0.5,length(xLabels)+0.5), ylim=c(1-0.5,length(yLabels)+0.5) , type="n", xlab="", ylab="", axes=FALSE)
    labelMatrix <- t(labelMatrix)
    xx <- rep(1:nrow(labelMatrix),ncol(labelMatrix))
    xx <- ((ncol(labelMatrix)+0.5)/ncol(labelMatrix) )*xx - 0.375
    yy <- as.numeric(as.vector(gl(ncol(labelMatrix),nrow(labelMatrix))))
    yy <- ((nrow(labelMatrix)+0.5)/nrow(labelMatrix) )*yy - 0.375
    zz <- as.vector(labelMatrix)
    stopifnot( length(xx) == length(yy) & length(xx) == length(zz) )
    text(xx,yy,zz,cex=labelMatrixTextScaling, adj=c(0,0))

  }

  
}# end of the function definition

