# /**
#
# \name{qualityParameters}
#
# \title{Calculation of quality characteristics for cDNA chip hybridisations}
#
# \alias{qualityParameters}
#
# \keyword{}
#
# \description{ } 
#
# \usage{ }
#
#
# \value{returns a list of results, i.e. a data frame \code{qualityParameters} containing
#   several scores for each hybridisation, a matrix \code{slideDistance} and
#   an integer \code{replicateSpots}, i.e. the number of spot replicates.
# 
#   The matrix \code{slideDistance} contains a distance measure (similarity) for each pair
#   of slides$_{ij}$, i.e. the median absolute deviation taken over all spots of the
#   difference of the "log-ratios" ("log-ratios": the difference of the expression values
#   in the two channels on the slide).
#
#   A brief summary of all parameters given the data frame \code{qualityParameters}:
#   \code{width}
#   a robust measure of the variance, i.e. the median absolute deviation
#   of the difference of the normalised channels taken over all spots
#
#   \code{medianDistance}
#   a robust measure for the typical distance (similarity) of one slide
#   with all other slides, i.e. the median of the "distances"
#   between slides (c.f. \code{slideDistance}))
#
#   \code{correlation}
#   of the expression values between the two channels of the slide taken over all spots
#
#   \code{meanIntensityCy3}
#   the mean taken over all spots of raw data channel Cy3
#
#   \code{meanIntensityCy5}
#   the mean taken over all spots of raw data channel Cy5
#
#   \code{meanIntensity}
#   mean taken over all spots of raw data of both channels,
#
#   \code{spotSimilarityCy3}
#   the standard deviation of all spot replicates for each unique identifier
#   of the normalised channel Cy3 is calculated and the
#   median standard deviation
#   is taken as spotSimilarityCy3; if duplicate spots have been detected, i.e.
#   \code{replicateSpots == 2}, the correlation is calculated instead
#
#   \code{spotSimilarityCy3}
#   the standard deviation of all spot replicates for each unique identifier
#   of the normalised channel Cy3 is calculated and the
#   median standard deviation
#   is taken as spotSimilarityCy3; if duplicate spots have been detected, i.e.
#   \code{replicateSpots == 2}, the correlation is calculated instead
#
#   \code{cy3vsAllCy3} and \code{cy5vsAllCy5}
#   the correlation between each channel is measured against the
#   averaged (median) channel over all hybridisations
#   (e.g. a virtual reference)
# }
#
#
# \arguments{
#  \item{arrayDataObject}{object of type \code{arrayData}; default: missing}
#  \item{exprSetArrayObject}{object of type \code{exprSetArray}; default: missing}
#  \item{spotIdentifier}{ character string, column name used in getSpotAttr(arrayDataObject);
#                         the column is used to determine spot replicas;
#                         default: "Name"}
#  \item{hybNameColumn}{ character string, column name used in getHybAttr(arrayDataObject);
#                        the column is used to extract the names of the hybridisations;
#                        if not available the order is taken instead;
#                        default:  "slideNumber"}
#  \item{identifiersToBeSkipped}{vector of character strings of spot identifiers
#                                to be excluded from calculations}
#  \item{resultFileName}{character string; results are stored in a file if supplied;
#                        default: missing}
#  \item{verbose}{logical; default \code{TRUE}}
# }
# 
# \details{}
#
# \seealso{\code{}}
#
#  \examples{
#      spotIdentifierVec <- c("A","A","Blank","B","B","Blank")
#      hybNames <- "H1"
#      R1 <- N1 <- c(1,1,9,2,2,10)
#      R2 <- N2 <- c(2,2,7,4,4,8)
#      rawDataIntensities <- array(0, dim=c(6,2,1))
#      rawDataIntensities[,1,] <- R1
#      rawDataIntensities[,2,] <- R2
#      dimnames(rawDataIntensities) <- list(NULL, c("Cy3","Cy5"), NULL)
#      spotattr=data.frame(Name=I(spotIdentifierVec))
#      hybattr=data.frame(slideNumber=I(hybNames))
#      arrayDataObject <- new("arrayData", intensities=rawDataIntensities, hybattr=hybattr, spotattr=spotattr)
#      indCy3=1
#      indCy5=2
#      channels <- matrix( c(indCy3,indCy5), nrow=length(indCy3), byrow=FALSE )
#      colnames(channels) <- c("Cy3","Cy5")
#      exprSetArrayObject <- new("exprSetArray", 	
#	exprs=matrix(c(R1,R2), nrow=6, byrow=FALSE), phenoData=	
#          new("phenoData", pData=data.frame(matrix(0,nrow=2,ncol=1)),
#              varLabels=list(rep("varLabel1",1))), channels=channels)		
#      Re1 <- qualityParameters(arrayDataObject=arrayDataObject, exprSetArrayObject=exprSetArrayObject, identifiersToBeSkipped= "Blank")
#      stopifnot(all.equal.numeric(as.numeric(Re1$qualityParameters["H1",c("correlation")]),c(1)))
#      stopifnot(Re1$replicateSpots==2)
#   \testonly{
#      spotIdentifierVec <- c("Blank","A","A","A","B","B","B")
#      R1 <- N1 <- c(10,1,1,1,2,2,2)
#      R2 <- N2 <- c(11,2,2,2,4,4,4)
#      rawDataIntensities <- array(0, dim=c(7,2,1))
#      rawDataIntensities[,1,] <- R1
#      rawDataIntensities[,2,] <- R2
#      dimnames(rawDataIntensities) <- list(NULL, c("Cy3","Cy5"), NULL)
#      spotattr=data.frame(Name=I(spotIdentifierVec))
#      hybattr=data.frame(slideNumber=I(hybNames))
#      arrayDataObject <- new("arrayData", intensities=rawDataIntensities, hybattr=hybattr, spotattr=spotattr)
#      indCy3=1
#      indCy5=2
#      channels <- matrix( c(indCy3,indCy5), nrow=length(indCy3), byrow=FALSE )
#      colnames(channels) <- c("Cy3","Cy5")
#      exprSetArrayObject <- new("exprSetArray", 	
#	exprs=matrix(c(R1,R2), nrow=7, byrow=FALSE), phenoData=	
#          new("phenoData", pData=data.frame(matrix(0,nrow=2,ncol=1)),
#              varLabels=list(rep("varLabel1",1))), channels=channels)		
#
#
#      Re2 <- qualityParameters(arrayDataObject=arrayDataObject, exprSetArrayObject=exprSetArrayObject, identifiersToBeSkipped= "Blank")
#
#      stopifnot(all.equal.numeric(as.numeric(Re2$qualityParameters["H1",c("correlation")]),c(1)))
#      stopifnot(Re2$replicateSpots==3)
#
#      hybNames <- c("H1","H2")
#      Xr1 <- Xn1 <- cbind(R1,R2)
#      Xr2 <- Xn2 <- cbind(R2,R1)
#      rawDataIntensities <- array(0, dim=c(7,2,2))
#      rawDataIntensities[,1,1:2] <- Xr1
#      rawDataIntensities[,2,1:2] <- Xr2
#      dimnames(rawDataIntensities) <- list(NULL, c("Cy3","Cy5"), NULL)
#      spotattr=data.frame(Name=I(spotIdentifierVec))
#      hybattr=data.frame(slideNumber=I(hybNames))
#      arrayDataObject <- new("arrayData", intensities=rawDataIntensities, hybattr=hybattr, spotattr=spotattr)
#      indCy3=1:2
#      indCy5=3:4
#      channels <- matrix( c(indCy3,indCy5), nrow=length(indCy3), byrow=FALSE )
#      colnames(channels) <- c("Cy3","Cy5")
#      exprSetArrayObject <- new("exprSetArray", 	
#	exprs=as.matrix(cbind(Xr1,Xr2)), phenoData=	
#          new("phenoData", pData=data.frame(matrix(0,nrow=4,ncol=1)),
#              varLabels=list(rep("varLabel1",1))), channels=channels)		
#
#
#      Re3 <- qualityParameters(arrayDataObject=arrayDataObject, exprSetArrayObject=exprSetArrayObject, identifiersToBeSkipped= "Blank")
#
#      stopifnot(all.equal.numeric(as.numeric(Re3$qualityParameters["H2",c("correlation")]),c(1)))
#      stopifnot(Re3$replicateSpots==3)
#   }
#
#	}
#
# \keyword{}
#
# \author{Andreas Buness <a.buness@dkfz.de>}
# */


qualityParameters <- function(arrayDataObject, exprSetArrayObject,
                              spotIdentifier = "Name",
                              hybNameColumn =  "slideNumber",
                              identifiersToBeSkipped=NA,
                              resultFileName,
                              verbose = TRUE
                              ){

  if( missing(exprSetArrayObject) ){
    
    warning(" exprSetArrayObject is missing in qualityParameters ")
    return()
    
  }
  
  if( missing(arrayDataObject) ){
    
    warning(" arrayDataObject is missing in qualityParameters ")
    return()
      
  }
  
  if( ! class(arrayDataObject) == "arrayData" ){
    warning(" arrayDataObject of wrong type in qualityParameters ")
    return(NULL)
  }
    
  if( ! class(exprSetArrayObject) == "exprSetArray" ){
    warning(" exprSetArrayObject of wrong type in qualityParameters ")
    return(NULL)
  }

  nChannelCy3 <- exprs(getExprSetCy3(exprSetArrayObject))
  nChannelCy5 <- exprs(getExprSetCy5(exprSetArrayObject))
  channelCy3 <- getIntensities(arrayDataObject)[,"Cy3",]
  channelCy5 <- getIntensities(arrayDataObject)[,"Cy5",]
    
  if( spotIdentifier %in% colnames(getSpotAttr(arrayDataObject)) ){
    spotIdentifierVec <- getSpotAttr(arrayDataObject)[,spotIdentifier]
  }else{
    warning(" spotIdentifier:",spotIdentifier," does not exist in colnames(getSpotAttr(arrayDataObject)) in qualityParameters")
    return(NULL)
  }
  if( hybNameColumn %in% colnames(getHybAttr(arrayDataObject)) ){
    hybNames <- getHybAttr(arrayDataObject)[, hybNameColumn]
  }else{
    warning(" hybNameColumn:",hybNameColumn," does not exist in colnames(getHybAttr(arrayDataObject))  in qualityParameters")
    hybNames <- as.character(1:dim(channelCy3)[2])
  }
  
  nChannelCy3 <- as.matrix(nChannelCy3)
  nChannelCy5 <- as.matrix(nChannelCy5)
  channelCy3 <- as.matrix(channelCy3)
  channelCy5 <- as.matrix(channelCy5)

  nrOfHybs <- length(hybNames)
  nrOfSpots <- length(spotIdentifierVec)

  dimVec <- c(nrOfSpots,nrOfHybs)

  if( any(dimVec != dim(nChannelCy3)) ||
      any(dimVec != dim(nChannelCy5)) ||
      any(dimVec != dim(channelCy3))  ||
      any(dimVec != dim(channelCy5))     ){
    
    warning(" inconsistent dimensions in qualityParameters ", call.=FALSE )
    return(NULL)
  }

  
  nonValidItems <- sapply(identifiersToBeSkipped, function(x) return(which( spotIdentifierVec == x)), simplify=TRUE )
  nonValidItems <- na.exclude(as.numeric(unlist(nonValidItems)))

  if( length(nonValidItems) > 0 ){

    if(verbose){cat(paste(" the number of excluded spots in qualityParameters is ",length(nonValidItems), "\n",sep=""))}

    spotIdentifierVec <- spotIdentifierVec[-nonValidItems]
    nChannelCy3 <- as.matrix(nChannelCy3[-nonValidItems,])
    nChannelCy5 <- as.matrix(nChannelCy5[-nonValidItems,])
    channelCy3 <- as.matrix(channelCy3[-nonValidItems,])
    channelCy5 <- as.matrix(channelCy5[-nonValidItems,])
  }

  replicateSpots <- NA
  spotSimCy3 <- numeric(length=nrOfHybs) * NA
  spotCy5 <- numeric(length=nrOfHybs) * NA
  names(spotSimCy3) <- hybNames
  names(spotCy5) <- hybNames
      
  # replicate spot correlation
  if( ! missing(spotIdentifierVec) ){

    if( ! any(duplicated(spotIdentifierVec)) ){
      # single spots
      # nothing to do

    }else if( max(table(spotIdentifierVec)) == 2 && min(table(spotIdentifierVec)) == 2 ){
      # solely duplicate spots
      # calculation of the correlation

      replicateSpots <- 2
      
      spotReplicas <- sapply(unique(spotIdentifierVec), function(x) return(which( spotIdentifierVec == x )) )
      
      for(i in 1:nrOfHybs){
        
        spotSimCy3[i] <- cor(nChannelCy3[spotReplicas[1,],i], nChannelCy3[spotReplicas[2,],i])
        spotCy5[i] <- cor(nChannelCy5[spotReplicas[1,],i], nChannelCy5[spotReplicas[2,],i])

      }
      
    }else if( max(table(spotIdentifierVec)) == min(table(spotIdentifierVec))
                         && max(table(spotIdentifierVec)) > 2 ){  
      # solely replicate spots of fixed number larger than two  
      # calculation of the mad of the variance

      replicateSpots <- max(table(spotIdentifierVec))
      spotReplicas <- sapply(unique(spotIdentifierVec), function(x) return(which( spotIdentifierVec == x )) )

      for(i in 1:nrOfHybs){
        spotSimCy3[i] <-  mad(apply(spotReplicas[,], 2, function(x) sd(nChannelCy3[x])))
        spotCy5[i] <-  mad(apply(spotReplicas[,], 2, function(x) sd(nChannelCy5[x])))
      }
        
    }else{

      # any mixture of single, duplicate and multiple spots
      # skipped so far

    }

  }else{# end of if not missing spotIdentifierVec
    warning(" no spot identifiers available in qualityParameters \n")
  }
    
  sigma <- matrix(NA, nrow=nrOfHybs, ncol=nrOfHybs)
  dimnames(sigma) <- list(hybNames,hybNames)
  medianSigma <- numeric(nrOfHybs) * NA
  names(medianSigma) <- hybNames
  sigmaVec <- numeric(nrOfHybs) * NA  # width
  names(sigmaVec) <- hybNames
  corVec <- numeric(nrOfHybs) * NA
  names(corVec) <- hybNames
  meanIntVecCy3 <- numeric(nrOfHybs) * NA
  names(meanIntVecCy3) <- hybNames
  meanIntVecCy5 <- numeric(nrOfHybs) * NA
  names(meanIntVecCy5) <- hybNames
  cy3CorWithAvCy3 <- numeric(nrOfHybs) * NA
  names(cy3CorWithAvCy3) <- hybNames
  cy5CorWidthAvCy5 <-  numeric(nrOfHybs) * NA
  names(cy5CorWidthAvCy5) <- hybNames
  
  # cross comparison - pre-calculations
  if( nrOfHybs > 1 ){
    for(i in 1:(nrOfHybs-1)){
      vecI <- nChannelCy3[,i]-nChannelCy5[,i]
      for(j in (i+1):nrOfHybs){
        vecJ <- nChannelCy3[,j]-nChannelCy5[,j]
        sigmaTmp <- mad(vecI - vecJ)
        sigma[i,j] <- sigma[j,i] <- sigmaTmp
      }
    }
  }

  for(i in 1:nrOfHybs){

    corVec[i] <- cor(nChannelCy3[,i],nChannelCy5[,i])
    sigmaVec[i] <- mad(nChannelCy3[,i]-nChannelCy5[,i])
    meanIntVecCy3[i] <- mean(channelCy3[,i])
    meanIntVecCy5[i] <- mean(channelCy5[,i])
    
    if( nrOfHybs > 1 ){
      medianSigma[i] <- median(sigma[i,-i])
    }
  }

  # channel comparison with "averaged" channel
  averageFunction <- median
  averagedChannelCy3 <- apply( nChannelCy3, 1, averageFunction)
  averagedChannelCy5 <- apply( nChannelCy5, 1, averageFunction)

  cy3CorWithAvCy3 <- apply( nChannelCy3, 2, function(x) cor(x,averagedChannelCy3))
  cy5CorWithAvCy5 <- apply( nChannelCy5, 2, function(x) cor(x,averagedChannelCy5))



  resultDF <- data.frame( hybridisation=I(hybNames),
                          width=sigmaVec,
                          medianDistance=medianSigma,
                          correlation=corVec,
                          meanIntensityCy3=meanIntVecCy3,
                          meanIntensityCy5=meanIntVecCy5,
                          meanIntensity= ( meanIntVecCy3 + meanIntVecCy5 )/2,
                          spotSimilarityCy3=spotSimCy3,
                          spotSimilarityCy5=spotCy5,
                          cy3vsAllCy3=cy3CorWithAvCy3,
                          cy5vsAllCy5=cy5CorWithAvCy5)
    
  
  if( ! missing(resultFileName) ){
    # write results to file
    qcTable <- resultDF
    write.table(qcTable, file=file.path(resultFileName), quote=FALSE, sep="\t",row.names=FALSE)
  }

  resultList <- list(qualityParameters=resultDF,
                     slideDistance = sigma,
                     replicateSpots = replicateSpots
                     )

  return(resultList)
  
}# end of function qualityParameters
